/* Non-font-backend font code. */

static inline int
ns_encode_char (int c, XChar2b *char2b, struct font_info *font_info,
                struct charset *charset, int *two_byte_p)
// ----------------------------------------------------------------------------
//23:  External (RIF); encode char to font
// ----------------------------------------------------------------------------
{
// unused w/font backend
    // our renderer passes unicode to the system, so no need to convert now
    *char2b = c;
    if (*char2b != c)
      add_to_log("ns_encode_char(): char value %d above 0xffff not supported.",
                 make_number(c), Qnil);
    return FONT_TYPE_UNKNOWN; // return value is used only by W32
}

static inline XCharStruct *
ns_per_char_metric (XFontStruct *font, XChar2b *char2b, int font_type)
// ----------------------------------------------------------------------------
//23:   compute metrics for one char and put in return struct
// ----------------------------------------------------------------------------
{
  xassert (font && char2b);

  return (font->per_char && *char2b < METRICS_CACHE_SIZE) ?
    font->per_char + XCHAR2B_BYTE2(char2b) : &(font->max_bounds);
}


struct font_info *
ns_load_font (struct frame *f, char *fontname, int size)
// ----------------------------------------------------------------------------
//  Load font named FONTNAME of the size SIZE for frame F, and return a
//  pointer to the structure font_info while allocating it dynamically.
//  If SIZE is 0, load any size of font.  If loading fails, return NULL.
// ----------------------------------------------------------------------------
{
  // placeholder -- unused w/font backend
  struct ns_display_info *dpyinfo = FRAME_NS_DISPLAY_INFO (f);
  int i;
  NSString *name;
  BOOL synthItal;

  NSTRACE(ns_load_font);

  //PENDING: shouldn't happen?
  if (fontname == NULL || !strncmp(fontname, "nil", 3))
      fontname = "Monaco";
  else if (fontname[0] == '-')
      fontname = ns_xlfd_to_fontname(fontname);

  synthItal = removeSynthItal(fontname, &name);

  if (size > 0)
    { 
      for (i = 0; i < dpyinfo->n_fonts; i++)
        if (dpyinfo->font_table[i].name
            && !strcmp (dpyinfo->font_table[i].name, fontname)
            && (dpyinfo->font_table[i].size == size))
          {
            struct font_info *fontInfo = dpyinfo->font_table + i;
            if (fontInfo)
                [((XFontStruct *)fontInfo->font)->nsfont retain];
            return (dpyinfo->font_table + i);
          }
    }

  if (size <= 0)
    {
      // try to get it out of frame params
        Lisp_Object tem = get_frame_param(f, Qfontsize);
        size = NILP(tem) ? 0 : XFASTINT(tem);
    }

  /* Load the font and add it to the table.  */
  {
    XFontStruct *font;
    struct font_info *fontp;
    NSFont *nsfont, *sfont;
    double expand, shrink, hshrink;
    float full_height, min_height, hd;
    char *realFontname = fontname;

    BLOCK_INPUT;

    nsfont = [NSFont fontWithName: name size: (float)size];

    if (!nsfont)
      {
//fprintf(stderr,"Warning: font '%s' not found; substituting default.\n", fontname);
        nsfont = [NSFont userFixedPitchFontOfSize: size];
        if (!nsfont)
          abort();
        realFontname = [[nsfont fontName] UTF8String];
      }
    size = [nsfont pointSize];
//NSLog(@"Loaded font '%@' of size %d w/ital angle %f.",nsfont, size,[nsfont italicAngle]);

    sfont = [nsfont screenFont];
    if (sfont == nil)
      sfont = nsfont;

    // set up ns_font (XFontStruct) structure
    font = (struct ns_font *)xmalloc(sizeof(struct ns_font));
    bzero (font, sizeof(struct ns_font));
#ifdef GNUSTEP
    font->nsfont = sfont;
#else
    font->nsfont = nsfont;
#endif
    [font->nsfont retain];
    font->name = (char *)xmalloc(strlen(realFontname)+1);
    bcopy(realFontname, font->name, strlen(realFontname)+1);
    font->ital = synthItal ||
     ([[NSFontManager sharedFontManager] traitsOfFont:nsfont]&NSItalicFontMask);

    // Set up metrics etc. in 'font' ("XfontStruct", defined in nsgui.h)
    // Some fonts return an unusually large max advance, so we only use it for
    // fonts that have wide characters.
    font->width = [sfont numberOfGlyphs] > 1000 ?
      [sfont maximumAdvancement].width : [sfont widthOfString: @"n"];

    full_height = [sfont boundingRectForFont].size.height;
    min_height = [sfont ascender] - [sfont descender];
    hd = full_height - min_height;

    if (!NUMBERP(ns_expand_space))
      error("No expand space defined");

    // expand = 1.0 means use full_height, 0.0 use min_height, negative shrink
    expand = XFLOATINT(ns_expand_space);

    if (expand < 0.0)
      {
        shrink = 1 + expand;
        hshrink = 1 + expand / 2.0;
        expand = 0.0;
      }
    else
      shrink = hshrink = 1.0;

    font->underpos = 2; //[sfont underlinePosition];//PENDING: often clipped out
    font->underwidth = [sfont underlineThickness];
    font->size = size;
    font->voffset = rint(hshrink * [sfont ascender] + expand * hd / 2);

    //23: max bounds
    font->max_bounds.ascent= rint(hshrink * [sfont ascender] + expand * hd / 2);
    font->max_bounds.descent= -rint(hshrink * [sfont descender] - expand *hd/2);
    font->height = font->max_bounds.ascent + font->max_bounds.descent;
    font->max_bounds.width = font->width;
    font->max_bounds.lbearing = 0;
    font->max_bounds.rbearing =
      font->width + (font->ital ? 0.2 * font->height : 0);
//if (font->ital) fprintf(stderr, "rb = %d, wd = %d\th = %.0f\n",font->max_bounds.rbearing,font->max_bounds.width,font->height);

    //23: Set up a char metrics cache that works more or less identically to
    //    those in other terms.  We store ASCII characters here.
    //    If these are all the same, we set cache null, which means use font
    //    max_bounds.  Unlike other terms, we cache these chars for "2-byte"
    //    fonts, since these will, at least for now, be used to render both
    //    ASCII and CJK or whatever else.
    if (![sfont isFixedPitch])
      {
        char c[2];
        c[1] = '\0';
        font->per_char = xmalloc(METRICS_CACHE_SIZE * sizeof(XCharStruct));
        for (i=0; i<METRICS_CACHE_SIZE; i++)
          {
            c[0] = i;
            //PENDING: find way to use boundingRectForGlyph and do this right
            NSString *s = [NSString stringWithCString: c];
            font->per_char[i] = font->max_bounds;
            font->per_char[i].width = [sfont widthOfString: s];
            font->per_char[i].rbearing =
              font->per_char[i].width + (font->ital ? 0.2 * font->height : 0);
          }
      }
    else
      {
        font->per_char = NULL;
      }

#ifdef COCOA
    // set up synthItal and the CG font
    font->synthItal = synthItal;
    {
      ATSFontRef atsFont = ATSFontFindFromPostScriptName
        ((CFStringRef)[nsfont fontName], kATSOptionFlagsDefault);

      if (atsFont == kATSFontRefUnspecified)
        {
          // see if we can get it by dropping italic (then synthesizing)
          atsFont = ATSFontFindFromPostScriptName((CFStringRef) [
              [[NSFontManager sharedFontManager]
                convertFont:nsfont toNotHaveTrait:NSItalicFontMask]
                                                                  fontName],
                                                  kATSOptionFlagsDefault);
          if (atsFont != kATSFontRefUnspecified)
          {
            font->synthItal = YES;
          }
          else
            {
              // last resort fallback
              atsFont = ATSFontFindFromPostScriptName
                ((CFStringRef)@"Monaco", kATSOptionFlagsDefault);
            }
        }
      font->cgfont = CGFontCreateWithPlatformFont((void*)&atsFont);
    }
#endif

    /* Allocate and fill in fontp structure.  */
    if (dpyinfo->font_table_size == 0)
      {
        dpyinfo->font_table_size = 16;
        dpyinfo->font_table
          = (struct font_info *) xmalloc (dpyinfo->font_table_size
                                          * sizeof (struct font_info));
      }
    else if (dpyinfo->n_fonts >= dpyinfo->font_table_size)
      {
        dpyinfo->font_table_size *= 2;
        dpyinfo->font_table
          = (struct font_info *) xrealloc (dpyinfo->font_table,
                                           (dpyinfo->font_table_size
                                            * sizeof (struct font_info)));
      }
    fontp = dpyinfo->font_table + dpyinfo->n_fonts;

    fontp->font = font;
    fontp->font_idx = dpyinfo->n_fonts;
    // name is the requested name, possibly including 'synthItal'
    fontp->name = (char *) xmalloc (strlen (fontname) + 1);
    bcopy (fontname, fontp->name, strlen (fontname) + 1);
    // full_name is the XLFD name
    fontp->full_name = xstrdup(ns_font_to_xlfd(nsfont));
    fontp->size = font->width;
    fontp->height = font->height;
    fontp->charset = -1; //PENDING: see w32
    fontp->space_width = [sfont widthOfString: @" "];
    fontp->average_width = font->width;
    fontp->default_ascent = 0;
    fontp->vertical_centering = 0;
    fontp->baseline_offset = 0; //PENDING: w32 uses 0
    fontp->relative_compose = 0;
    //PENDING: w32 sets '4' for shift-jis fonts..
    fontp->encoding_type = FONT_ENCODING_NOT_DECIDED;
    fontp->font_encoder = NULL;
    //PENDING: in other terms, encoder set later by x_find_ccl_program...

    dpyinfo->n_fonts++;

    UNBLOCK_INPUT;

    return fontp;
  }
}


void
ns_unload_font (struct frame *f, XFontStruct *font)
// ----------------------------------------------------------------------------
//   Called to unload a font (when a frame is dropped, for example).
// ----------------------------------------------------------------------------
{
  if (!font || font == ((XFontStruct *) FACE_DEFAULT))
    return;

  if (font->nsfont)
    {
//NSLog(@"ns_unload_font: %@\n",font->nsfont);
      [font->nsfont release];
      font->nsfont = nil;
#ifdef COCOA
      CGFontRelease(font->cgfont);
#endif
      xfree(font->name);
      if (font->per_char)
        xfree(font->per_char);
      xfree(font);
    }
}


BOOL
removeSynthItal(char *str, NSString **ptr)
// ----------------------------------------------------------------------------
// Utility: find & chop off "-synthItal" from a font name, return YES if found
// ----------------------------------------------------------------------------
{
  NSRange range;

  if (str == NULL)
      return NO;

  *ptr = [NSString stringWithUTF8String: str];

  range = [*ptr rangeOfString: @"-synthItal"
                      options: NSBackwardsSearch | NSCaseInsensitiveSearch];
  if (range.location + range.length == [*ptr length])
    {
      range.location = 0;
      range.length = [*ptr length] - range.length;
      *ptr = [*ptr substringWithRange: range];
      return YES;
    }
  else
    return NO;
}


//PENDING: not used when font backend active (and doesn't seem to be
//         used even when it isn't)
Lisp_Object
x_new_fontset (struct frame *f, Lisp_Object fontsetname)
{
  fprintf(stderr, "x_set_font called x_new_fontset: '%s'\n",
          XSTRING(fontsetname)->data);
//      ns_new_font(f,
//                  [[[NSFont userFixedPitchFontOfSize: -1] fontName] UTF8String]);
    return build_string
        ([[[NSFont userFixedPitchFontOfSize: -1] fontName] UTF8String]);
}


Lisp_Object
ns_new_font (struct frame *f, char *fontname)
// ----------------------------------------------------------------------------
//    Called to adjust frame size when frame main font changed.
// ----------------------------------------------------------------------------
{
  struct font_info *fontp = FS_LOAD_FONT(f, fontname);
  int fw;
  NSSize sz;

  NSTRACE(ns_new_font);

  if (!fontp)
    return Qnil;

  if (FRAME_FONT (f) == (XFontStruct *) (fontp->font))
    {
      Lisp_Object tem = get_frame_param(f, Qfontsize);
      int size = NILP(tem) ? 0 : XFASTINT(tem);

      /* This font is already set in frame F.  There's nothing more to do. */
      fprintf(stderr, "Returning because of same font!\n");
      fprintf(stderr, "ns_font_size: %d, vs current size %d\n",size,
              ((XFontStruct *)fontp->font)->size);
      return build_string (fontp->full_name);
    }

  FRAME_FONT (f) = (XFontStruct *) (fontp->font);
  fw = fontp->average_width;
  FRAME_COLUMN_WIDTH (f) = fw;
  FRAME_SPACE_WIDTH (f) = fontp->space_width;
  FRAME_LINE_HEIGHT (f) = FONT_HEIGHT (FRAME_FONT (f));

  sz.width = fw;
  sz.height = FRAME_LINE_HEIGHT(f);
  [[FRAME_NS_VIEW(f) window] setResizeIncrements: sz];

  compute_fringe_widths (f, 1);

  /* Compute the scroll bar width in character columns.  */
  if (f->config_scroll_bar_width > 0)
    f->config_scroll_bar_cols = (f->config_scroll_bar_width + fw-1) / fw;
  else
    f->config_scroll_bar_cols = (NS_SCROLL_BAR_WIDTH_DEFAULT + fw-1) / fw;

    /* Now make the frame display the given font.  */
  if (FRAME_NS_VIEW (f) != 0)
    {
      // do this whether or not height changed: width may have changed
      x_set_window_size (f, 0, FRAME_COLS (f), FRAME_LINES (f));
    }
  else

  return build_string (fontp->full_name);
}


struct font_info *
ns_query_font (struct frame *f, char *fontname)
// ----------------------------------------------------------------------------
//  External (hook): for fontset impl: Return font_info for name, or null.
// ----------------------------------------------------------------------------
{
  int i;
  struct ns_display_info *dpyinfo = ns_display_list;

  NSTRACE(ns_query_font);
  for (i = 0; i < dpyinfo->n_fonts; i++)
    if (!strcmp (dpyinfo->font_table[i].name, fontname)
	|| !strcmp (dpyinfo->font_table[i].full_name, fontname))
      return (dpyinfo->font_table + i);
  return NULL;
}
